import torch.nn as nn
from torch.nn.utils import weight_norm
import numpy as np

from coopertunes.hparams import MelGanHParams


def weights_init(m):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find("BatchNorm2d") != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)


def WNConv1d(*args, **kwargs):
    return weight_norm(nn.Conv1d(*args, **kwargs))


def WNConvTranspose1d(*args, **kwargs):
    return weight_norm(nn.ConvTranspose1d(*args, **kwargs))


class ResnetBlock(nn.Module):
    def __init__(self, dim, dilation=1):
        super().__init__()
        self.block = nn.Sequential(
            nn.LeakyReLU(0.2),
            nn.ReflectionPad1d(dilation),
            WNConv1d(dim, dim, kernel_size=3, dilation=dilation),
            nn.LeakyReLU(0.2),
            WNConv1d(dim, dim, kernel_size=1),
        )
        self.shortcut = WNConv1d(dim, dim, kernel_size=1)

    def forward(self, x):
        return self.shortcut(x) + self.block(x)


class MelGanGenerator(nn.Module):
    """
    Generating raw audio from mel spectrogram with GAN generator.
    """

    def __init__(self, hparams: MelGanHParams):
        super().__init__()
        ratios = hparams.generator_ratios
        self.hop_length = np.prod(ratios)
        mult = int(2 ** len(ratios))

        model = [
            nn.ReflectionPad1d(3),
            WNConv1d(hparams.n_mel_channels, mult *
                     hparams.ngf, kernel_size=7, padding=0),
        ]

        # Upsample to raw audio scale
        for r in ratios:
            model += [
                nn.LeakyReLU(0.2),
                WNConvTranspose1d(
                    mult * hparams.ngf,
                    mult * hparams.ngf // 2,
                    kernel_size=r * 2,
                    stride=r,
                    padding=r // 2 + r % 2,
                    output_padding=r % 2,
                ),
            ]

            for j in range(hparams.n_residual_layers):
                model += [ResnetBlock(mult * hparams.ngf //
                                      2, dilation=3 ** j)]

            mult //= 2

        model += [
            nn.LeakyReLU(0.2),
            nn.ReflectionPad1d(3),
            WNConv1d(hparams.ngf, 1, kernel_size=7, padding=0),
            nn.Tanh(),
        ]

        self.model = nn.Sequential(*model)
        self.apply(weights_init)

    def forward(self, x):
        return self.model(x)

    def inference(self, x):
        return self.model(x)


class MelGanNLayerDiscriminator(nn.Module):
    def __init__(self, hparams: MelGanHParams):
        super().__init__()
        model = nn.ModuleDict()

        model["layer_0"] = nn.Sequential(
            nn.ReflectionPad1d(7),
            WNConv1d(1, hparams.ndf, kernel_size=15),
            nn.LeakyReLU(0.2, True),
        )

        nf = hparams.ndf
        stride = hparams.downsamp_factor
        for n in range(1, hparams.n_layers_D + 1):
            nf_prev = nf
            nf = min(nf * stride, 1024)

            model[f"layer_{n}"] = nn.Sequential(
                WNConv1d(
                    nf_prev,
                    nf,
                    kernel_size=stride * 10 + 1,
                    stride=stride,
                    padding=stride * 5,
                    groups=nf_prev // 4,
                ),
                nn.LeakyReLU(0.2, True),
            )

        nf = min(nf * 2, 1024)
        model[f"layer_{hparams.n_layers_D + 1}"] = nn.Sequential(
            WNConv1d(nf_prev, nf, kernel_size=5, stride=1, padding=2),
            nn.LeakyReLU(0.2, True),
        )

        model[f"layer_{hparams.n_layers_D + 2}"] = WNConv1d(
            nf, 1, kernel_size=3, stride=1, padding=1
        )

        self.model = model

    def forward(self, x):
        results = []
        for _, layer in self.model.items():
            x = layer(x)
            results.append(x)
        return results


class MelGanDiscriminator(nn.Module):
    def __init__(self, hparams: MelGanHParams):
        super().__init__()
        self.model = nn.ModuleDict()
        for i in range(hparams.num_D):
            self.model[f"disc_{i}"] = MelGanNLayerDiscriminator(hparams)

        self.downsample = nn.AvgPool1d(
            4, stride=2, padding=1, count_include_pad=False)
        self.apply(weights_init)

    def forward(self, x):
        results = []
        for _, disc in self.model.items():
            results.append(disc(x))
            x = self.downsample(x)
        return results
